# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
from transformers import AutoProcessor, MllamaConfig
from transformers.models.mllama.configuration_mllama import MllamaTextConfig, MllamaVisionConfig

from nemo import lightning as nl
from nemo.collections import vlm


def split_qkv_weight(qkv_weight, model_config):
    """Split attention qkv from nemo to hf format"""
    hidden_size = model_config.hidden_size
    head_num = model_config.num_attention_heads
    num_query_groups = model_config.num_query_groups or head_num
    head_size = model_config.kv_channels or (hidden_size // head_num)
    heads_per_group = head_num // num_query_groups
    qkv_weight = qkv_weight.reshape(-1, head_size, hidden_size)
    q_weight = torch.empty((head_num, head_size, hidden_size), device=qkv_weight.device)
    k_weight = torch.empty((num_query_groups, head_size, hidden_size), device=qkv_weight.device)
    v_weight = torch.empty((num_query_groups, head_size, hidden_size), device=qkv_weight.device)

    qkv_index = 0
    for i in range(num_query_groups):
        q_weight[i * heads_per_group : (i + 1) * heads_per_group, :, :] = qkv_weight[
            qkv_index : qkv_index + heads_per_group, :, :
        ]
        qkv_index += heads_per_group
        k_weight[i, :, :] = qkv_weight[qkv_index, :, :]
        qkv_index += 1
        v_weight[i, :, :] = qkv_weight[qkv_index, :, :]
        qkv_index += 1

    return [('q_proj', q_weight), ('k_proj', k_weight), ('v_proj', v_weight)]


def split_kv_weight(kv_weight, model_config):
    """Split cross attention qkv from nemo to hf format"""
    hidden_size = model_config.hidden_size
    head_num = model_config.num_attention_heads
    num_query_groups = model_config.num_query_groups or head_num
    head_size = model_config.kv_channels or (hidden_size // head_num)
    kv_weight = kv_weight.reshape(-1, head_size, hidden_size)
    k_weight = torch.empty((num_query_groups, head_size, hidden_size), device=kv_weight.device)
    v_weight = torch.empty((num_query_groups, head_size, hidden_size), device=kv_weight.device)

    kv_index = 0
    for i in range(num_query_groups):
        k_weight[i, :, :] = kv_weight[kv_index, :, :]
        kv_index += 1
        v_weight[i, :, :] = kv_weight[kv_index, :, :]
        kv_index += 1

    return [('k_proj', k_weight), ('v_proj', v_weight)]


def split_gate_weight(gate_weight):
    """Split linear fc to gate"""
    gate_weight = torch.chunk(gate_weight, 2, axis=0)

    return [('gate_proj', gate_weight[0]), ('up_proj', gate_weight[1])]


def convert_mllama_config(source_vision, source_text):
    """Convert nemo mllama config to hf config"""
    vision_config = MllamaVisionConfig(
        num_hidden_layers=source_vision.num_layers,
        hidden_size=source_vision.hidden_size,
        attention_heads=source_vision.num_attention_heads,
        image_size=source_vision.vision_chunk_size,
        max_num_tiles=source_vision.vision_max_num_chunks,
        torch_dtype="bfloat16",
    )

    cross_attention_layers = [
        x + i for i, x in enumerate(source_text._init_fusion_schedule(source_text.num_cross_attention_layers))
    ]
    text_config = MllamaTextConfig(
        rope_theta=source_text.rotary_base,
        num_hidden_layers=source_text.num_layers + source_text.num_cross_attention_layers,
        cross_attention_layers=cross_attention_layers,
        hidden_size=source_text.hidden_size,
        intermediate_size=source_text.ffn_hidden_size,
        num_attention_heads=source_text.num_attention_heads,
        num_key_value_heads=source_text.num_query_groups,
        vocab_size=source_text.vocab_size,
        rope_scaling={
            "factor": 8.0,
            "high_freq_factor": 4.0,
            "low_freq_factor": 1.0,
            "original_max_position_embeddings": 8192,
            "rope_type": "llama3",
        },
        eos_token_id=[128001, 128008, 128009],
        torch_dtype="bfloat16",
    )

    return MllamaConfig(vision_config, text_config, torch_dtype="bfloat16")


def convert_mllama_nemo_to_hf(checkpoint_path, processor_name):
    """Convert nemo mllama to hf state dict and config"""
    processor = AutoProcessor.from_pretrained(processor_name)

    strategy = nl.MegatronStrategy(
        tensor_model_parallel_size=1,
        ckpt_load_optimizer=False,
        ckpt_save_optimizer=False,
    )
    trainer = nl.Trainer(
        devices=1,
        max_steps=1000,
        accelerator="gpu",
        strategy=strategy,
        plugins=nl.MegatronMixedPrecision(precision="bf16-mixed"),
        val_check_interval=1000,
        limit_val_batches=50,
    )

    fabric = trainer.to_fabric()

    tokenizer = processor.tokenizer
    model = vlm.MLlamaModel(vlm.MLlamaConfig11BInstruct(), tokenizer=tokenizer)
    config = model.config
    vision_model_config = config.vision_model_config
    language_model_config = config.language_model_config
    model = fabric.load_model(checkpoint_path, model)
    model = model.module.module.module.module

    state_dict = model.state_dict()
    del model

    v = "vision_model.vision_encoder"
    key_map = [
        ("vision_model.class_embedding", f"{v}.class_embedding"),
        ("vision_model.gated_positional_embedding.embedding", f"{v}.positional_embedding"),
        (
            "vision_model.gated_positional_embedding.tile_embedding.weight",
            f"{v}.gated_tile_positional_embedding.weight",
        ),
        ("vision_model.gated_positional_embedding.gate", f"{v}.gated_positional_embedding_gate"),
        ("vision_model.layernorm_post.bias", f"{v}.ln_post.bias"),
        ("vision_model.layernorm_post.weight", f"{v}.ln_post.weight"),
        ("vision_model.layernorm_pre.bias", f"{v}.ln_pre.bias"),
        ("vision_model.layernorm_pre.weight", f"{v}.ln_pre.weight"),
        ("vision_model.post_tile_positional_embedding.embedding.weight", f"{v}.post_tile_pos_embed.embedding.weight"),
        ("vision_model.post_tile_positional_embedding.gate", f"{v}.post_tile_pos_embed.gate"),
        ("vision_model.pre_tile_positional_embedding.embedding.weight", f"{v}.pre_tile_pos_embed.embedding.weight"),
        ("vision_model.pre_tile_positional_embedding.gate", f"{v}.pre_tile_pos_embed.gate"),
        ("multi_modal_projector.bias", "vision_model.vision_projection.encoder.bias"),
        ("multi_modal_projector.weight", "vision_model.vision_projection.encoder.weight"),
        ("language_model.model.norm.weight", "language_model.decoder.final_layernorm.weight"),
        ("language_model.lm_head.weight", "language_model.output_layer.weight"),
    ]

    for i in range(vision_model_config.num_layers):
        key_map.extend(
            [
                (
                    f"vision_model.transformer.layers.{i}.self_attn.o_proj.weight",
                    f"{v}.transformer.layers.{i}.self_attention.linear_proj.weight",
                ),
                (
                    f"vision_model.transformer.layers.{i}.input_layernorm.bias",
                    f"{v}.transformer.layers.{i}.input_layernorm.bias",
                ),
                (
                    f"vision_model.transformer.layers.{i}.input_layernorm.weight",
                    f"{v}.transformer.layers.{i}.input_layernorm.weight",
                ),
                (
                    f"vision_model.transformer.layers.{i}.post_attention_layernorm.bias",
                    f"{v}.transformer.layers.{i}.pre_mlp_layernorm.bias",
                ),
                (
                    f"vision_model.transformer.layers.{i}.post_attention_layernorm.weight",
                    f"{v}.transformer.layers.{i}.pre_mlp_layernorm.weight",
                ),
                (
                    f"vision_model.transformer.layers.{i}.mlp.fc1.bias",
                    f"{v}.transformer.layers.{i}.mlp.linear_fc1.bias",
                ),
                (
                    f"vision_model.transformer.layers.{i}.mlp.fc1.weight",
                    f"{v}.transformer.layers.{i}.mlp.linear_fc1.weight",
                ),
                (
                    f"vision_model.transformer.layers.{i}.mlp.fc2.bias",
                    f"{v}.transformer.layers.{i}.mlp.linear_fc2.bias",
                ),
                (
                    f"vision_model.transformer.layers.{i}.mlp.fc2.weight",
                    f"{v}.transformer.layers.{i}.mlp.linear_fc2.weight",
                ),
            ]
        )

    for i in range(vision_model_config.num_global_layers):
        key_map.extend(
            [
                (
                    f"vision_model.global_transformer.layers.{i}.self_attn.o_proj.weight",
                    f"{v}.global_transformer.layers.{i}.self_attention.linear_proj.weight",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.gate_attn",
                    f"{v}.global_transformer.layers.{i}.gate_attn",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.gate_ffn",
                    f"{v}.global_transformer.layers.{i}.gate_ffn",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.input_layernorm.bias",
                    f"{v}.global_transformer.layers.{i}.input_layernorm.bias",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.input_layernorm.weight",
                    f"{v}.global_transformer.layers.{i}.input_layernorm.weight",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.post_attention_layernorm.bias",
                    f"{v}.global_transformer.layers.{i}.pre_mlp_layernorm.bias",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.post_attention_layernorm.weight",
                    f"{v}.global_transformer.layers.{i}.pre_mlp_layernorm.weight",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.mlp.fc1.bias",
                    f"{v}.global_transformer.layers.{i}.mlp.linear_fc1.bias",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.mlp.fc1.weight",
                    f"{v}.global_transformer.layers.{i}.mlp.linear_fc1.weight",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.mlp.fc2.bias",
                    f"{v}.global_transformer.layers.{i}.mlp.linear_fc2.bias",
                ),
                (
                    f"vision_model.global_transformer.layers.{i}.mlp.fc2.weight",
                    f"{v}.global_transformer.layers.{i}.mlp.linear_fc2.weight",
                ),
            ]
        )

    cross_attention_frequency = language_model_config.num_layers // language_model_config.num_cross_attention_layers
    toal_num_layer = language_model_config.num_layers + language_model_config.num_cross_attention_layers
    prefix = "language_model.decoder"
    for i in range(toal_num_layer):
        cross_num = (i - 3) // (cross_attention_frequency + 1)
        if (i - 3) % (cross_attention_frequency + 1) == 0:
            xattn_index = cross_num * cross_attention_frequency + 3
            key_map.extend(
                [
                    (
                        f"language_model.model.layers.{i}.cross_attn.o_proj.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.cross_attention.linear_proj.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.cross_attn.q_proj.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.cross_attention.linear_q.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.cross_attn.k_norm.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.cross_attention.k_layernorm.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.input_layernorm.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.cross_attention.linear_q.layer_norm_weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.cross_attn.q_norm.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.cross_attention.q_layernorm.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.post_attention_layernorm.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.mlp.linear_fc1.layer_norm_weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.mlp.down_proj.weight",
                        f"{prefix}.xattn_layers.{xattn_index}.mlp.linear_fc2.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.cross_attn_attn_gate",
                        f"{prefix}.xattn_layers.{xattn_index}.gate_attn",
                    ),
                    (
                        f"language_model.model.layers.{i}.cross_attn_mlp_gate",
                        f"{prefix}.xattn_layers.{xattn_index}.gate_ffn",
                    ),
                ]
            )
        else:
            attn_index = i - cross_num - 1
            key_map.extend(
                [
                    (
                        f"language_model.model.layers.{i}.self_attn.o_proj.weight",
                        f"{prefix}.layers.{attn_index}.self_attention.linear_proj.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.post_attention_layernorm.weight",
                        f"{prefix}.layers.{attn_index}.mlp.linear_fc1.layer_norm_weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.mlp.down_proj.weight",
                        f"{prefix}.layers.{attn_index}.mlp.linear_fc2.weight",
                    ),
                    (
                        f"language_model.model.layers.{i}.input_layernorm.weight",
                        f"{prefix}.layers.{attn_index}.self_attention.linear_qkv.layer_norm_weight",
                    ),
                ]
            )

    new_state_dict = {}
    for new_key, old_key in key_map:
        new_state_dict[new_key] = state_dict[old_key]

    def convert_vision_qkv_weight(state_dict, vision_model_config):
        hidden_size = vision_model_config.hidden_size

        new_state_dict = {}
        for i in range(vision_model_config.num_layers):
            qkv_weights = state_dict[
                f"vision_model.vision_encoder.transformer.layers.{i}.self_attention.linear_qkv.weight"
            ]

            for name, weight in split_qkv_weight(qkv_weights, vision_model_config):
                new_key = f'vision_model.transformer.layers.{i}.self_attn.{name}.weight'
                new_state_dict[new_key] = weight.reshape(-1, hidden_size)

        for i in range(vision_model_config.num_global_layers):
            qkv_weights = state_dict[
                f"vision_model.vision_encoder.global_transformer.layers.{i}.self_attention.linear_qkv.weight"
            ]

            for name, weight in split_qkv_weight(qkv_weights, vision_model_config):
                new_key = f'vision_model.global_transformer.layers.{i}.self_attn.{name}.weight'
                new_state_dict[new_key] = weight.reshape(-1, hidden_size)

        return new_state_dict

    def convert_patch_embeding(state_dict):
        conv1_weight = state_dict["vision_model.vision_encoder.conv1._linear.weight"]
        return {"vision_model.patch_embedding.weight": conv1_weight.reshape(conv1_weight.shape[0], 3, 14, 14)}

    def convert_language_qkv_weight(state_dict, language_model_config):
        hidden_size = language_model_config.hidden_size
        new_state_dict = {}
        for i in range(toal_num_layer):
            cross_num = (i - 3) // (cross_attention_frequency + 1)
            if (i - 3) % (cross_attention_frequency + 1) == 0:
                xattn_index = cross_num * cross_attention_frequency + 3
                kv_weights = state_dict[f"{prefix}.xattn_layers.{xattn_index}.cross_attention.linear_kv.weight"]
                for name, weight in split_kv_weight(kv_weights, language_model_config):
                    new_key = f"language_model.model.layers.{i}.cross_attn.{name}.weight"
                    new_state_dict[new_key] = weight.reshape(-1, hidden_size)
            else:
                attn_index = i - cross_num - 1
                qkv_weights = state_dict[f"{prefix}.layers.{attn_index}.self_attention.linear_qkv.weight"]
                for name, weight in split_qkv_weight(qkv_weights, language_model_config):
                    new_key = f"language_model.model.layers.{i}.self_attn.{name}.weight"
                    new_state_dict[new_key] = weight.reshape(-1, hidden_size)

        return new_state_dict

    def convert_gate(state_dict):
        new_state_dict = {}
        for i in range(toal_num_layer):
            cross_num = (i - 3) // (cross_attention_frequency + 1)
            if (i - 3) % (cross_attention_frequency + 1) == 0:
                xattn_index = cross_num * cross_attention_frequency + 3
                gate_weight = state_dict[f"{prefix}.xattn_layers.{xattn_index}.mlp.linear_fc1.weight"]
            else:
                attn_index = i - cross_num - 1
                gate_weight = state_dict[f"{prefix}.layers.{attn_index}.mlp.linear_fc1.weight"]

            for name, weight in split_gate_weight(gate_weight):
                new_key = f"language_model.model.layers.{i}.mlp.{name}.weight"
                new_state_dict[new_key] = weight

        return new_state_dict

    def convert_embedding(state_dict):
        word_embeddings = state_dict["language_model.embedding.word_embeddings.weight"]
        learnable_embedding = state_dict["language_model.learnable_embedding.weight"]

        return {"language_model.model.embed_tokens.weight": torch.cat((word_embeddings, learnable_embedding), dim=0)}

    new_state_dict.update(convert_vision_qkv_weight(state_dict, vision_model_config))
    new_state_dict.update(convert_patch_embeding(state_dict))
    new_state_dict.update(convert_language_qkv_weight(state_dict, language_model_config))
    new_state_dict.update(convert_gate(state_dict))
    new_state_dict.update(convert_embedding(state_dict))

    return new_state_dict, convert_mllama_config(vision_model_config, language_model_config)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional

import lightning.pytorch as pl

try:
    from megatron.core.distributed import finalize_model_grads
    from megatron.core.optimizer import OptimizerConfig
    from megatron.core.utils import get_model_config

    HAVE_MEGATRON_CORE = True

except (ImportError, ModuleNotFoundError):

    OptimizerConfig = object
    HAVE_MEGATRON_CORE = False

from torch.optim import Optimizer

from nemo.lightning._strategy_lib import setup_megatron_optimizer
from nemo.lightning.megatron_parallel import MegatronParallel
from nemo.lightning.pytorch.optim.base import LRSchedulerModule, OptimizerModule


class MegatronOptimizerModule(OptimizerModule):
    """A OptimizerModule for the megatron optimizers.

    Attributes:
        config (OptimizerConfig): Configuration for the optimizer.
        no_weight_decay_cond (Optional[Callable]): Condition for no weight decay.
        scale_lr_cond (Optional[Callable]): Condition for scaling learning rate.
        lr_mult (float): Learning rate multiplier.

    Example::

        config = OptimizerConfig(...)
        lr_scheduler = MyLRSchedulerModule(...)
        optimizer_module = MegatronOptimizerModule(config, lr_scheduler)

    Methods:
        setup(model): Sets up the optimizer.
        optimizers(model): Defines the optimizers.
    """

    def __init__(
        self,
        config: OptimizerConfig,
        lr_scheduler: Optional[LRSchedulerModule] = None,
        no_weight_decay_cond: Optional[Callable] = None,
        scale_lr_cond: Optional[Callable] = None,
        lr_mult: float = 1.0,
    ):
        """Initializes the MegatronOptimizerModule.

        Args:
            config (OptimizerConfig): Configuration for the optimizer.
            lr_scheduler (Optional[LRSchedulerModule]): The learning rate scheduler module.
            no_weight_decay_cond (Optional[Callable]): Condition for no weight decay.
            scale_lr_cond (Optional[Callable]): Condition for scaling learning rate.
            lr_mult (float): Learning rate multiplier.
        """

        super().__init__(lr_scheduler=lr_scheduler)
        self.config = config
        self.no_weight_decay_cond = no_weight_decay_cond
        self.scale_lr_cond = scale_lr_cond
        self.lr_mult = lr_mult

    def on_fit_start(self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"):
        """We will add the finalize_model_grads function to the model config.

        Args:
            model: The model for which the optimizer is being set up.
        """

        def finalize_model_grads_func(*args, **kwargs):
            return self.finalize_model_grads(*args, **kwargs)

        get_model_config(pl_module).finalize_model_grads_func = finalize_model_grads_func

    def optimizers(self, model: MegatronParallel) -> List[Optimizer]:
        """Defines the optimizers.

        Args:
            model (MegatronParallel): The model for which the optimizers are being defined.

        Returns:
            List[Optimizer]: The list of optimizers.

        Raises:
            ValueError: If the model is not an instance of MegatronParallel.
        """

        if not isinstance(model, MegatronParallel):
            raise ValueError("Model must be an instance of MegatronParallel")

        optimizer = setup_megatron_optimizer(
            model,
            self.config,
            no_weight_decay_cond=self.no_weight_decay_cond,
            scale_lr_cond=self.scale_lr_cond,
            lr_mult=self.lr_mult,
        )

        return [optimizer]

    def finalize_model_grads(self, *args, **kwargs):
        """Return function to finalize the model gradients."""
        return finalize_model_grads(*args, **kwargs)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import nemo_run as run
import pytest
import torch

from nemo.collections.llm.api import finetune, pretrain
from nemo.collections.llm.gpt.data.mock import MockDataModule
from nemo.collections.llm.gpt.data.squad import SquadDataModule
from nemo.collections.llm.gpt.model.llama import Llama31Config405B, LlamaModel
from nemo.collections.llm.peft.lora import LoRA
from nemo.collections.llm.recipes import llama31_405b
from nemo.lightning import Trainer
from nemo.lightning.pytorch.callbacks.garbage_collection import GarbageCollectionCallback


class TestLlama31_405B:
    @pytest.fixture(scope="class")
    def recipe_module(self):
        return llama31_405b

    def test_model(self, recipe_module):
        model_config = recipe_module.model()
        assert isinstance(model_config, run.Config)
        assert model_config.__fn_or_cls__ == LlamaModel
        assert isinstance(model_config.config, run.Config)
        assert model_config.config.__fn_or_cls__ == Llama31Config405B
        assert model_config.config.seq_length == 8192

    def test_trainer(self, recipe_module):
        trainer_config = recipe_module.trainer()
        assert isinstance(trainer_config, run.Config)
        assert trainer_config.__fn_or_cls__ == Trainer
        assert trainer_config.accelerator == "gpu"
        assert trainer_config.devices == 8
        assert trainer_config.num_nodes == 64  # Much larger than 70B
        assert trainer_config.max_steps == 1168251

        # Check strategy configuration
        assert isinstance(trainer_config.strategy, run.Config)
        assert trainer_config.strategy.__fn_or_cls__.__name__ == "MegatronStrategy"
        assert trainer_config.strategy.tensor_model_parallel_size == 8
        assert trainer_config.strategy.pipeline_model_parallel_size == 8
        assert trainer_config.strategy.pipeline_dtype == torch.bfloat16
        assert trainer_config.strategy.virtual_pipeline_model_parallel_size == 2
        assert trainer_config.strategy.context_parallel_size == 4
        assert trainer_config.strategy.sequence_parallel is True
        assert trainer_config.strategy.account_for_embedding_in_pipeline_split is True
        assert trainer_config.strategy.account_for_loss_in_pipeline_split is True

    def test_pretrain_recipe(self, recipe_module):
        recipe = recipe_module.pretrain_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == pretrain
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == LlamaModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == MockDataModule
        assert recipe.data.seq_length == 8192
        assert recipe.data.global_batch_size == 512
        assert recipe.data.micro_batch_size == 1

    def test_finetune_recipe(self, recipe_module):
        recipe = recipe_module.finetune_recipe()
        assert isinstance(recipe, run.Partial)
        assert recipe.__fn_or_cls__ == finetune
        assert isinstance(recipe.model, run.Config)
        assert recipe.model.__fn_or_cls__ == LlamaModel
        assert isinstance(recipe.trainer, run.Config)
        assert recipe.trainer.__fn_or_cls__ == Trainer
        assert isinstance(recipe.data, run.Config)
        assert recipe.data.__fn_or_cls__ == SquadDataModule
        assert recipe.data.seq_length == 2048
        assert recipe.data.global_batch_size == 6  # Different from other models
        assert isinstance(recipe.peft, run.Config)
        assert recipe.peft.__fn_or_cls__ == LoRA
        assert recipe.peft.dim == 16
        assert recipe.peft.alpha == 32
        assert recipe.optim.config.lr == 1e-4

        # Check pipeline split settings
        assert recipe.trainer.strategy.account_for_embedding_in_pipeline_split is True
        assert recipe.trainer.strategy.account_for_loss_in_pipeline_split is True

    def test_finetune_recipe_without_peft(self, recipe_module):
        recipe = recipe_module.finetune_recipe(peft_scheme=None)
        assert recipe.trainer.strategy.tensor_model_parallel_size == 8
        assert recipe.trainer.strategy.pipeline_model_parallel_size == 14
        assert recipe.data.global_batch_size == 6
        assert recipe.optim.config.lr == 5e-6
        assert not hasattr(recipe, 'peft') or recipe.peft is None

    def test_pretrain_performance_optimizations(self, recipe_module):
        recipe = recipe_module.pretrain_performance_optimizations(recipe_module.pretrain_recipe())
        assert recipe.trainer.plugins.grad_reduce_in_fp32 is False

        # Check callbacks
        assert any(
            isinstance(cb, run.Config) and cb.__fn_or_cls__ == GarbageCollectionCallback
            for cb in recipe.trainer.callbacks
        )
        comm_overlap_cb = next(
            cb
            for cb in recipe.trainer.callbacks
            if isinstance(cb, run.Config) and cb.__fn_or_cls__.__name__ == "MegatronCommOverlapCallback"
        )
        assert comm_overlap_cb.tp_comm_overlap is True
        assert comm_overlap_cb.defer_embedding_wgrad_compute is True
        assert comm_overlap_cb.wgrad_deferral_limit == 50
        assert comm_overlap_cb.overlap_param_gather_with_optimizer_step is False

    def test_finetune_performance_optimizations(self, recipe_module):
        recipe = recipe_module.finetune_recipe(performance_mode=True, peft_scheme=None)
        assert recipe.trainer.strategy.tensor_model_parallel_size == 8
        assert recipe.trainer.strategy.pipeline_model_parallel_size == 14
        assert recipe.trainer.strategy.sequence_parallel is True
        assert recipe.trainer.plugins.grad_reduce_in_fp32 is False

        # Check DDP settings
        assert recipe.trainer.strategy.ddp.grad_reduce_in_fp32 is False
        assert recipe.trainer.strategy.ddp.overlap_grad_reduce is True
        assert recipe.trainer.strategy.ddp.overlap_param_gather is True
        assert recipe.trainer.strategy.ddp.average_in_collective is True

    def test_finetune_performance_optimizations_with_peft(self, recipe_module):
        recipe = recipe_module.finetune_recipe(performance_mode=True, peft_scheme='lora')
        assert recipe.peft.target_modules == ['linear_qkv']
        assert recipe.trainer.strategy.tensor_model_parallel_size == 4
        assert recipe.trainer.strategy.pipeline_model_parallel_size == 4
        assert recipe.trainer.strategy.virtual_pipeline_model_parallel_size == 4
        assert recipe.trainer.strategy.sequence_parallel is True



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""HuggingFace model serialization support for NeMo's configuration system.

This module provides integration between NeMo's configuration system and HuggingFace's
pretrained models. It enables automatic serialization and deserialization of HuggingFace
models within NeMo's configuration framework.

The integration works by:
1. Detecting HuggingFace models through their characteristic methods (save_pretrained/from_pretrained)
2. Converting them to Fiddle configurations that preserve the model's class and path
3. Providing an artifact handler (HFAutoArtifact) that manages the actual model files

Example:
    ```python
    from transformers import AutoModel
    
    # This model will be automatically handled by the HFAutoArtifact system
    model = AutoModel.from_pretrained("bert-base-uncased")
    
    # When serialized, the model files will be saved to the artifacts directory
    # When deserialized, the model will be loaded from the saved files
    ```
"""

import contextlib
import inspect
import threading
from pathlib import Path

import fiddle as fdl

from nemo.lightning.io.artifact import Artifact
from nemo.lightning.io.to_config import to_config

_local = threading.local()


class HFAutoArtifact(Artifact):
    """Artifact handler for HuggingFace pretrained model/processor/tokenizer/etc..

    This handler manages the serialization and deserialization of HuggingFace models
    by utilizing their save_pretrained/from_pretrained methods. It saves models to
    an 'artifacts' subdirectory within the specified path.
    """

    def dump(self, instance, value: Path, absolute_dir: Path, relative_dir: Path) -> Path:
        """Save a HuggingFace model to disk.

        Args:
            instance: The HuggingFace model instance to save
            value: Original path value (unused)
            absolute_dir: Absolute path to the save directory
            relative_dir: Relative path from the config file to the save directory

        Returns:
            str: The relative path to the saved model artifacts
        """
        instance.save_pretrained(Path(absolute_dir) / "artifacts")
        return "./" + str(Path(relative_dir) / "artifacts")

    def load(self, path: Path) -> Path:
        """Return the path to load a HuggingFace model.

        Args:
            path: Path to the saved model artifacts

        Returns:
            Path: The same path, to be used with from_pretrained
        """
        return path


@contextlib.contextmanager
def from_pretrained_kwargs(**kwargs):
    """Context manager for passing additional kwargs to from_pretrained.

    Args:
        **kwargs: Keyword arguments to pass to from_pretrained

    Example:
        with from_pretrained_kwargs(trust_remote_code=True):
            io.load_context("path/to/checkpoint")
    """
    if not hasattr(_local, "kwargs"):
        _local.kwargs = {}
    previous = _local.kwargs.copy()
    _local.kwargs.update(kwargs)
    try:
        yield
    finally:
        _local.kwargs = previous


def from_pretrained(auto_cls, pretrained_model_name_or_path="dummy"):
    """Factory function for loading HuggingFace pretrained models.

    This function is used as the serialization target for HuggingFace models.
    When deserialized, it will recreate the model using its from_pretrained method.

    Args:
        auto_cls: The HuggingFace model class (e.g., AutoModel, AutoTokenizer)
        pretrained_model_name_or_path: Path to the saved model or model identifier

    Returns:
        The loaded HuggingFace model
    """
    kwargs = getattr(_local, "kwargs", {})
    return auto_cls.from_pretrained(pretrained_model_name_or_path, **kwargs)


@to_config.register(
    lambda v: not inspect.isclass(v)
    and getattr(v, "__module__", "").startswith("transformers")
    and hasattr(v, "save_pretrained")
    and hasattr(v, "from_pretrained")
)
def handle_hf_pretrained(value):
    """Convert a HuggingFace model instance to a Fiddle configuration.

    This handler detects HuggingFace model instances by checking for the presence
    of save_pretrained and from_pretrained methods. It converts them to a Fiddle
    configuration that will recreate the model using from_pretrained.

    Args:
        value: A HuggingFace model instance

    Returns:
        fdl.Config: A Fiddle configuration that will recreate the model
    """
    return fdl.Config(
        from_pretrained,
        auto_cls=value.__class__,
        pretrained_model_name_or_path="dummy",
    )


# Register the HFAutoArtifact handler for the pretrained_model_name_or_path parameter
from_pretrained.__io_artifacts__ = [HFAutoArtifact("pretrained_model_name_or_path")]



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

from omegaconf import DictConfig

from nemo.collections.asr.parts.utils import adapter_utils
from nemo.collections.tts.modules.aligner import AlignmentEncoder
from nemo.collections.tts.modules.fastpitch import TemporalPredictor
from nemo.collections.tts.modules.transformer import FFTransformerDecoder, FFTransformerEncoder
from nemo.core.classes import adapter_mixins


class FFTransformerDecoderAdapter(FFTransformerDecoder, adapter_mixins.AdapterModuleMixin):
    """ Inherit from FFTransformerDecoder and add support for adapter"""

    def add_adapter(self, name: str, cfg: dict):
        cfg = self._update_adapter_cfg_input_dim(cfg)
        for fft_layer in self.layers:  # type: adapter_mixins.AdapterModuleMixin
            fft_layer.add_adapter(name, cfg)

    def is_adapter_available(self) -> bool:
        return any([FFT_layer.is_adapter_available() for FFT_layer in self.layers])

    def set_enabled_adapters(self, name: Optional[str] = None, enabled: bool = True):
        for FFT_layer in self.layers:  # type: adapter_mixins.AdapterModuleMixin
            FFT_layer.set_enabled_adapters(name=name, enabled=enabled)

    def get_enabled_adapters(self) -> List[str]:
        names = set([])
        for FFT_layer in self.layers:  # type: adapter_mixins.AdapterModuleMixin
            names.update(FFT_layer.get_enabled_adapters())

        names = sorted(list(names))
        return names

    def _update_adapter_cfg_input_dim(self, cfg: DictConfig):
        cfg = adapter_utils.update_adapter_cfg_input_dim(self, cfg, module_dim=self.d_model)
        return cfg


class FFTransformerEncoderAdapter(
    FFTransformerDecoderAdapter, FFTransformerEncoder, adapter_mixins.AdapterModuleMixin
):
    """ Inherit from FFTransformerEncoder and add support for adapter"""

    pass


class AlignmentEncoderAdapter(AlignmentEncoder, adapter_mixins.AdapterModuleMixin):
    """ Inherit from AlignmentEncoder and add support for adapter"""

    def add_adapter(self, name: str, cfg: dict):

        for i, conv_layer in enumerate(self.key_proj):
            if i % 2 == 0:
                cfg = self._update_adapter_cfg_input_dim(cfg, conv_layer.conv.out_channels)
                conv_layer.add_adapter(name, cfg)

        for i, conv_layer in enumerate(self.query_proj):
            if i % 2 == 0:
                cfg = self._update_adapter_cfg_input_dim(cfg, conv_layer.conv.out_channels)
                conv_layer.add_adapter(name, cfg)

    def is_adapter_available(self) -> bool:
        return any(
            [conv_layer.is_adapter_available() for i, conv_layer in enumerate(self.key_proj) if i % 2 == 0]
            + [conv_layer.is_adapter_available() for i, conv_layer in enumerate(self.query_proj) if i % 2 == 0]
        )

    def set_enabled_adapters(self, name: Optional[str] = None, enabled: bool = True):
        for i, conv_layer in enumerate(self.key_proj):
            if i % 2 == 0:
                conv_layer.set_enabled_adapters(name=name, enabled=enabled)
        for i, conv_layer in enumerate(self.query_proj):
            if i % 2 == 0:
                conv_layer.set_enabled_adapters(name=name, enabled=enabled)

    def get_enabled_adapters(self) -> List[str]:
        names = set([])
        for i, conv_layer in enumerate(self.key_proj):
            if i % 2 == 0:
                names.update(conv_layer.get_enabled_adapters())
        for i, conv_layer in enumerate(self.query_proj):
            if i % 2 == 0:
                names.update(conv_layer.get_enabled_adapters())

        names = sorted(list(names))
        return names

    def _update_adapter_cfg_input_dim(self, cfg: DictConfig, module_dim: int):
        cfg = adapter_utils.update_adapter_cfg_input_dim(self, cfg, module_dim=module_dim)
        return cfg


class TemporalPredictorAdapter(TemporalPredictor, adapter_mixins.AdapterModuleMixin):
    """ Inherit from TemporalPredictor and add support for adapter"""

    def add_adapter(self, name: str, cfg: dict):
        cfg = self._update_adapter_cfg_input_dim(cfg)
        for conv_layer in self.layers:  # type: adapter_mixins.AdapterModuleMixin
            conv_layer.add_adapter(name, cfg)

    def is_adapter_available(self) -> bool:
        return any([conv_layer.is_adapter_available() for conv_layer in self.layers])

    def set_enabled_adapters(self, name: Optional[str] = None, enabled: bool = True):
        for conv_layer in self.layers:  # type: adapter_mixins.AdapterModuleMixin
            conv_layer.set_enabled_adapters(name=name, enabled=enabled)

    def get_enabled_adapters(self) -> List[str]:
        names = set([])
        for conv_layer in self.layers:  # type: adapter_mixins.AdapterModuleMixin
            names.update(conv_layer.get_enabled_adapters())

        names = sorted(list(names))
        return names

    def _update_adapter_cfg_input_dim(self, cfg: DictConfig):
        cfg = adapter_utils.update_adapter_cfg_input_dim(self, cfg, module_dim=self.filter_size)
        return cfg


"""Register any additional information"""
if adapter_mixins.get_registered_adapter(FFTransformerEncoder) is None:
    adapter_mixins.register_adapter(base_class=FFTransformerEncoder, adapter_class=FFTransformerEncoderAdapter)

if adapter_mixins.get_registered_adapter(FFTransformerDecoder) is None:
    adapter_mixins.register_adapter(base_class=FFTransformerDecoder, adapter_class=FFTransformerDecoderAdapter)

if adapter_mixins.get_registered_adapter(AlignmentEncoder) is None:
    adapter_mixins.register_adapter(base_class=AlignmentEncoder, adapter_class=AlignmentEncoderAdapter)

if adapter_mixins.get_registered_adapter(TemporalPredictor) is None:
    adapter_mixins.register_adapter(base_class=TemporalPredictor, adapter_class=TemporalPredictorAdapter)



# Copyright (c) 2022, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import (
    PrecisionRecallDisplay,
    RocCurveDisplay,
    average_precision_score,
    log_loss,
    precision_recall_curve,
    roc_auc_score,
    roc_curve,
)


def auc_roc(y_true: Union[List[int], np.ndarray], y_score: Union[List[float], np.ndarray]) -> float:
    """Compute Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores.

    Note: If only one class is present in y_true, 0.5 is returned.
    """
    y_true = np.array(y_true)
    y_score = np.array(y_score)
    assert len(y_true) == len(y_score)
    assert np.all(y_true >= 0) and np.all(y_true <= 1)
    if np.all(y_true == 0) or np.all(y_true == 1):
        return 0.5
    return roc_auc_score(y_true, y_score)


def auc_pr(y_true: Union[List[int], np.ndarray], y_score: Union[List[float], np.ndarray]) -> float:
    """Compute Area Under the Precision-Recall Curve (PR AUC) from prediction scores.

    Note: If only regatives are present in y_true, 0.0 is returned.
    """
    y_true = np.array(y_true)
    y_score = np.array(y_score)
    assert len(y_true) == len(y_score)
    assert np.all(y_true >= 0) and np.all(y_true <= 1)
    if np.all(y_true == 0):
        return 0.0
    return average_precision_score(y_true, y_score)


def auc_nt(y_true: Union[List[int], np.ndarray], y_score: Union[List[float], np.ndarray]) -> float:
    """Compute Area Under the Negative Predictive Value vs. True Negative Rate Curve (NT AUC) from prediction scores.

    This metric can be thought of as a PR AUC in which errors are treated as positives.

    Note: If only positives are present in y_true, 0.0 is returned.
    """
    y_true = np.array(y_true)
    y_score = np.array(y_score)
    assert len(y_true) == len(y_score)
    assert np.all(y_true >= 0) and np.all(y_true <= 1)
    if np.all(y_true == 1):
        return 0.0
    return average_precision_score(1 - y_true, 1 - y_score)


def nce(y_true: Union[List[int], np.ndarray], y_score: Union[List[float], np.ndarray]) -> float:
    """Compute Normalized Cross Entropy (NCE) from prediction scores. Also known as the Normalized Mutual Information.

    NCE measures how close the correct prediction scores are to one and the incorrect prediction scores are to zero.
    Negative NCE values indicate that the classifier performs worse than the setting all prediction scores
    as the proportion of correct predictions.

    Note: If only one class is present in y_true, 0.5 is returned.
    """
    y_true = np.array(y_true)
    y_score = np.array(y_score)
    assert len(y_true) == len(y_score)
    assert np.all(y_true >= 0) and np.all(y_true <= 1)
    if np.all(y_true == 0) or np.all(y_true == 1):
        return -math.inf
    p = y_true.mean()
    eps = 1e-15
    Hp = -(math.log(p + eps) * p + math.log(1 - p + eps) * (1 - p))
    return (Hp - log_loss(y_true, y_score)) / Hp


def ece(
    y_true: Union[List[int], np.ndarray],
    y_score: Union[List[float], np.ndarray],
    n_bins: int = 100,
    return_curve: bool = False,
) -> Union[float, Tuple[float, Tuple[List[int], List[float]]]]:
    """Compute Expected Calibration Error (ECE) from prediction scores.

    ECE measures how close the correct prediction scores are to one and the incorrect prediction scores are to zero.
    ECE ranges from zero to one with the best value zero (the lower the value, the better).
    """
    y_true = np.array(y_true)
    y_score = np.array(y_score)
    assert len(y_true) == len(y_score)
    assert np.all(y_true >= 0) and np.all(y_true <= 1)
    py = np.array([1 - y_score, y_score]).T
    acc, conf = np.zeros(n_bins), np.zeros(n_bins)
    Bm = np.zeros(n_bins)
    ece_curve = []
    thresholds = []
    for m in range(n_bins):
        a, b = m / n_bins, (m + 1) / n_bins
        threshold = (a + b) / 2
        thresholds.append(threshold)
        py_index = (py.T[1] >= threshold).astype(int)
        py_value = py[np.arange(len(py_index)), py_index]
        bin_range = ((py_value > a) & (py_value <= b)).nonzero()[0]
        Bm[m] = len(bin_range)
        if Bm[m] > 0:
            acc[m] = (py_index[bin_range] == y_true[bin_range]).sum() / Bm[m]
            conf[m] = py_value[bin_range].sum() / Bm[m]
        ece_curve.append(Bm[m] * np.abs(acc[m] - conf[m]))
    ece = sum(ece_curve) / sum(Bm)
    if return_curve:
        return ece, (thresholds, ece_curve)
    else:
        return ece


def auc_yc(
    y_true: Union[List[int], np.ndarray],
    y_score: Union[List[float], np.ndarray],
    n_bins: int = 100,
    return_std_maximum: bool = False,
    return_curve: bool = False,
) -> Union[
    float,
    Tuple[float, Tuple[List[int], List[float]]],
    Tuple[float, float, float],
    Tuple[float, float, float, Tuple[List[int], List[float]]],
]:
    """Compute Area Under the Youden's Curve (YC AUC) from prediction scores.

    YC AUC represents the rate of the effective threshold range.

    If return_std_maximum is set to True, std and maximum values of the Youden's Curve are returned with the AUC.

    Note: If only one class is present in y_true, zeroes are returned for every entity.
    """
    y_true = np.array(y_true)
    y_score = np.array(y_score)
    thresholds = np.linspace(0, 1, n_bins + 1)
    assert len(y_true) == len(y_score)
    assert np.all(y_true >= 0) and np.all(y_true <= 1)
    if np.all(y_true == 0) or np.all(y_true == 1):
        if return_std_maximum and return_curve:
            return 0.0, 0.0, 0.0, (thresholds, np.zeros(len(thresholds)))
        elif return_std_maximum:
            return 0.0, 0.0, 0.0
        elif return_curve:
            return 0.0, (thresholds, np.zeros(len(thresholds)))
        else:
            return 0.0
    mask_correct = y_true == 1
    count_correct = max(len(mask_correct.nonzero()[0]), 1)
    count_incorrect = max(len(y_true) - count_correct, 1)
    y_score_correct = y_score[mask_correct]
    y_score_incorrect = y_score[~mask_correct]
    yc = []
    for threshold in thresholds:
        tnr = len((y_score_incorrect < threshold).nonzero()[0]) / count_incorrect
        fnr = len((y_score_correct < threshold).nonzero()[0]) / count_correct
        yc.append(abs(tnr - fnr))
    yc = np.array(yc)
    if return_std_maximum and return_curve:
        return yc.mean(), yc.std(), yc.max(), (thresholds, yc)
    elif return_std_maximum:
        return yc.mean(), yc.std(), yc.max()
    elif return_curve:
        return yc.mean(), (thresholds, yc)
    else:
        return yc.mean()


def save_confidence_hist(y_score: Union[List[float], np.ndarray], plot_dir: Union[str, Path], name: str = "hist"):
    os.makedirs(plot_dir, exist_ok=True)
    plt.hist(np.array(y_score), 50, range=(0, 1))
    plt.title(name)
    plt.xlabel("Confidence score")
    plt.ylabel("Count")
    plt.savefig(Path(plot_dir) / Path(name + ".png"), dpi=300)
    plt.clf()


def save_roc_curve(
    y_true: Union[List[int], np.ndarray],
    y_score: Union[List[float], np.ndarray],
    plot_dir: Union[str, Path],
    name: str = "roc",
):
    assert len(y_true) == len(y_score)
    os.makedirs(plot_dir, exist_ok=True)
    fpr, tpr, _ = roc_curve(1 - np.array(y_true), 1 - np.array(y_score))
    RocCurveDisplay(fpr=fpr, tpr=tpr).plot()
    plt.title(name)
    plt.savefig(Path(plot_dir) / Path(name + ".png"), dpi=300)
    plt.clf()


def save_pr_curve(
    y_true: Union[List[int], np.ndarray],
    y_score: Union[List[float], np.ndarray],
    plot_dir: Union[str, Path],
    name: str = "pr",
):
    assert len(y_true) == len(y_score)
    os.makedirs(plot_dir, exist_ok=True)
    precision, recall, _ = precision_recall_curve(np.array(y_true), np.array(y_score))
    PrecisionRecallDisplay(precision=precision, recall=recall).plot()
    plt.title(name)
    plt.savefig(Path(plot_dir) / Path(name + ".png"), dpi=300)
    plt.clf()


def save_nt_curve(
    y_true: Union[List[int], np.ndarray],
    y_score: Union[List[float], np.ndarray],
    plot_dir: Union[str, Path],
    name: str = "nt",
):
    assert len(y_true) == len(y_score)
    os.makedirs(plot_dir, exist_ok=True)
    precision, recall, _ = precision_recall_curve(1 - np.array(y_true), 1 - np.array(y_score))
    PrecisionRecallDisplay(precision=precision, recall=recall).plot()
    plt.title(name)
    plt.savefig(Path(plot_dir) / Path(name + ".png"), dpi=300)
    plt.clf()


def save_custom_confidence_curve(
    thresholds: Union[List[float], np.ndarray],
    values: Union[List[float], np.ndarray],
    plot_dir: Union[str, Path],
    name: str = "my_awesome_curve",
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
):
    assert len(thresholds) == len(values)
    os.makedirs(plot_dir, exist_ok=True)
    plt.plot(thresholds, values)
    plt.xlim([0, 1])
    plt.ylim([0, 1])
    plt.title(name)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    plt.savefig(Path(plot_dir) / Path(name + ".png"), dpi=300)
    plt.clf()



# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import deque
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import NamedTuple, Optional

import numpy as np
import sentencepiece as spm
import torch
from lightning.pytorch import Trainer
from omegaconf import MISSING, DictConfig, OmegaConf

from nemo.collections.asr.parts.context_biasing.context_graph_universal import ContextGraph, ContextState
from nemo.collections.asr.parts.submodules.ngram_lm import DEFAULT_TOKEN_OFFSET, NGramGPULanguageModel
from nemo.collections.common.tokenizers import AggregateTokenizer
from nemo.collections.common.tokenizers.tokenizer_spec import TokenizerSpec
from nemo.utils import logging
from nemo.utils.exceptions import NeMoBaseException


@dataclass
class PhraseItem:
    phrase: str  # phrase itself
    lang: str  # per-phrase language (for aggregate tokenizer)
    # custom weight can be further added


@dataclass
class BoostingTreeModelConfig:
    """
    Boosting tree model config
    """

    model_path: Optional[str] = None  # The path to builded '.nemo' boosting tree model
    key_phrases_file: Optional[str] = None  # The path to the context-biasing list file (one phrase per line)
    key_phrases_list: Optional[list[str]] = (
        None  # The list of context-biasing phrases ['word1', 'word2', 'word3', ...]
    )
    # The list of context-biasing phrases with custom options:
    # [PhraseItem("word1", lang="en"), PhraseItem("frase dos", lang="es"), ...]
    # in CLI: key_phrase_items_list='[{phrase:"word1",lang:en},{phrase:"frase dos",lang:es}]'
    key_phrase_items_list: list[PhraseItem] | None = None
    context_score: float = 1.0  # The score for each arc transition in the context graph
    depth_scaling: float = (
        2.0  # The scaling factor for the depth of the context graph (2.0 for CTC, RNN-T and TDT, 1.0 for Canary)
    )
    unk_score: float = (
        0.0  # The score for unknown tokens (tokens that are not presented in the beginning of context-biasing phrases)
    )
    final_eos_score: float = (
        1.0  # The score for eos token after detected end of context phrase to prevent hallucination for AED models
    )
    score_per_phrase: float = 0.0  # Custom score for each phrase in the context graph
    source_lang: str = "en"  # The source language of the context-biasing phrases (for aggregate tokenizer)
    use_triton: bool = True  # Whether to use Triton for inference.
    uniform_weights: bool = False  # Whether to use uniform weights for the context-biasing tree as in Icefall
    use_bpe_dropout: bool = False  # Whether to use BPE dropout for generating alternative transcriptions
    num_of_transcriptions: int = (
        5  # The number of alternative transcriptions to generate for each context-biasing phrase
    )
    bpe_alpha: float = 0.3  # The alpha parameter for BPE dropout

    @staticmethod
    def is_empty(cfg: "BoostingTreeModelConfig") -> bool:
        return (
            cfg.model_path is None
            and cfg.key_phrases_file is None
            and (not cfg.key_phrases_list)
            and (not cfg.key_phrase_items_list)
        )


class TBranch(NamedTuple):
    """Structure (tuple) to represent a branch in the boosting tree"""

    symbol: int  # token id
    start_node: ContextState  # start node of the branch
    next_node: ContextState  # next node of the branch


@dataclass
class BoostingTreeStorage:
    """
    NumPy-based storage for suffix tree (weighted acceptor) for phrase boosting
    """

    num_states_max: InitVar[int]
    num_arcs_max: InitVar[int]

    vocab_size: int
    max_order: int

    arcs: np.ndarray = field(init=False)
    states: np.ndarray = field(init=False)

    _node_cache: dict[int, int] = field(default_factory=dict)

    unk_score: float = 0.0
    final_eos_score: float = 0.0
    num_states: int = 0
    num_arcs: int = 0
    start_state: int = 0
    bos_state: int = 0

    def __post_init__(self, num_states_max: int, num_arcs_max: int):
        if max(num_states_max, num_arcs_max) < np.iinfo(np.int32).max:
            int_np_dtype = np.int32
        else:
            int_np_dtype = np.int64
        self.arcs = np.zeros(
            [num_arcs_max],
            dtype=[("from", int_np_dtype), ("to", int_np_dtype), ("ilabel", int_np_dtype), ("weight", np.float32)],
        )
        self.states = np.zeros(
            [num_states_max],
            dtype=[
                ("arcs_start", int_np_dtype),
                ("arcs_end", int_np_dtype),
                ("order", int_np_dtype),
                ("backoff_to", int_np_dtype),
                ("backoff_w", np.float32),
                ("final", np.float32),
            ],
        )
        self.states["final"] = self.final_eos_score
        self._node_cache[0] = 0
        self.separate_bos_state = False

    def _add_tbranches_first_order(self, tbranches: list):
        """Add all first order tbranches to the model (similar with unigrams for N-Gram LM)"""

        tbranches = sorted(tbranches, key=lambda x: (x.start_node.id, x.symbol))

        self.num_states = 1
        self.num_arcs = 0
        # state: start_arcs, end_arcs, order, backoff_to, backoff_weight
        self.states[self.start_state] = (0, self.vocab_size, 1, self.start_state, 0.0, 0.0)
        added_symbols = set()
        num_vocab_labels = 0
        for tbranch in tbranches:
            ilabel = tbranch.symbol
            assert ilabel < self.vocab_size
            arc_id = ilabel
            added_symbols.add(ilabel)
            next_state = self.num_states
            self.num_states += 1
            self.arcs[arc_id] = (self.start_state, next_state, ilabel, tbranch.next_node.token_score)
            self.num_arcs += 1

            if tbranch.next_node.is_end:
                # we do not penalize transitions from final nodes in case of non-uniform weights
                backoff_weight = 0.0
            else:
                backoff_weight = tbranch.next_node.fail.node_score - tbranch.next_node.node_score

            # state order
            self.states[next_state] = (
                0,
                0,
                self.states[self.start_state]["order"] + 1,
                self.start_state,
                backoff_weight,
                self.final_eos_score if tbranch.next_node.is_end else 0.0,
            )
            num_vocab_labels += 1
            self._node_cache[tbranch.next_node.id] = next_state

        for ilabel in range(self.vocab_size):
            if ilabel not in added_symbols:
                self.arcs[ilabel] = (self.start_state, self.start_state, ilabel, self.unk_score)
                self.num_arcs += 1

    def _add_tbranches_next_order(self, tbranches: list):
        """Add tbranches for the order > 1; should be called after adding first order tokens (unigrams), using increasing order"""
        tbranches = sorted(tbranches, key=lambda x: (x.start_node.id, x.symbol))

        for tbranch in tbranches:
            ilabel = tbranch.symbol
            from_state = self._node_cache[tbranch.start_node.id]
            assert ilabel < self.vocab_size
            backoff_state = self._node_cache[tbranch.next_node.fail.id]

            if tbranch.next_node.is_end and not self.uniform_weights:
                # we do not penalize transitions from final nodes in case of non-uniform weights
                backoff_weight = 0.0
            else:
                backoff_weight = tbranch.next_node.fail.node_score - tbranch.next_node.node_score

            arc_id = self.num_arcs
            next_state = self.num_states
            self.num_arcs += 1
            self.num_states += 1
            token_score = tbranch.next_node.token_score
            if self.uniform_weights and tbranch.next_node.is_end:
                token_score += tbranch.next_node.node_score

            self.arcs[arc_id] = (from_state, next_state, ilabel, token_score)

            self.states[next_state] = (
                0,
                0,
                self.states[from_state]["order"] + 1,
                backoff_state,
                backoff_weight,
                self.final_eos_score if tbranch.next_node.is_end else 0.0,
            )

            self._node_cache[tbranch.next_node.id] = next_state

            if self.states[from_state]["arcs_start"] == 0:
                self.states[from_state]["arcs_start"] = arc_id
                self.states[from_state]["arcs_end"] = arc_id + 1
            else:
                assert self.states[from_state]["arcs_end"] == arc_id
                self.states[from_state]["arcs_end"] = arc_id + 1

    def _start_adding_tbranches_for_order(self, order: int):
        """Prepare for adding tbranches for the given order: initialize temporary storage"""
        self._start_arcs = self.num_arcs
        self._cur_order = order
        self._tbranches = []
        self._tbranches_cnt = 0

    def _end_adding_tbranches_for_order(self, order: int):
        """Finish adding tbranches for the given order"""
        if order == 1:
            assert len(self._tbranches) == self._tbranches_cnt
            self._add_tbranches_first_order(tbranches=self._tbranches)
            self._tbranches = None
            self._tbranches_cnt = 0
        else:
            assert len(self._tbranches) == self._tbranches_cnt
            self._add_tbranches_next_order(tbranches=self._tbranches)
            self._tbranches = None
            self._tbranches_cnt = 0

    def sanity_check(self):
        """Sanity check for the model"""
        assert (self.arcs["ilabel"][: self.num_arcs] < self.vocab_size).all()
        assert (self.arcs["ilabel"][: self.num_arcs] >= 0).all()


@dataclass
class BoostingTreeConfig:
    """
    N-Gram LM Config
    """

    num_states: int = MISSING
    num_arcs: int = MISSING
    max_order: int = MISSING
    vocab_size: int = MISSING
    separate_bos_state: bool = False
    use_triton: bool | None = None


class GPUBoostingTreeModel(NGramGPULanguageModel):
    """
    GPU-accelerated boosting tree supporting batched queries.
    Fast implementation for parallel queries for full vocabulary.
    Supports autograd (differentiable weights).
    """

    START_STATE = 0

    def __init__(
        self,
        cfg: DictConfig,
        trainer: Trainer = None,
    ):
        """
        Stubs for constructor that does not initialize the structure.
        This constructor can be useful when storing/loading module using native torch serialization mechanism
        instead of directly reading ARPA model -> converting to Torch, which can be slow for large N-Gram models
        (of several GBs).

        Args:
            cfg:
                num_states: number of states in graph
                num_arcs: number of arcs (transitions) in graph
                max_order: maximum order of n-gram LM (maximum possible nubmer of transitions without backoffs)
                vocab_size: vocabulary size (existing vocabulary units in LM; should not include blank etc.)
                separate_bos_state: separate Begin-of-Sentence state (default: True - for n-gram LM)
                use_triton: allow using Triton implementation;
                    None (default) means "auto" (used if available), True means forced mode
                    (will crash if Triton is unavailable)
            trainer: Lightning trainer (optional)
        """
        super().__init__(cfg=cfg, trainer=trainer)
        self.bos_state = self.START_STATE  # Always START_STATE for gpu boosting tree

    @classmethod
    def _read_context_graph(
        cls,
        context_graph: ContextGraph,
    ) -> tuple[dict[int, int], list[TBranch]]:
        """
        Read context-biasing tree from python structure and return branches in TBranch format.

        Args:
            context_graph: python context-biasing graph
        """

        seen = set()
        queue = deque()
        queue.append(context_graph.root)
        seen.add(0)
        order2cnt = {}
        tbranches_list = []

        # read context graph tree in breadth-first order to add branches for boosting tree generation
        while len(queue):
            current_node = queue.popleft()
            for token, node in current_node.next.items():
                if node.id not in seen:
                    tbranches_list.append(TBranch(symbol=token, start_node=current_node, next_node=node))
                    order2cnt[node.level] = order2cnt.get(node.level, 0) + 1
                    queue.append(node)

        return order2cnt, tbranches_list

    @classmethod
    def from_context_graph(
        cls,
        context_graph: ContextGraph,
        vocab_size: int,
        unk_score: float = 0.0,
        final_eos_score: float = 0.0,
        use_triton: bool | None = None,
        uniform_weights: bool | None = None,
    ) -> "GPUBoostingTreeModel":
        """
        Constructor from Icefall context graph (dict-based tree).

        Args:
            context_graph: context-biasing graph
            vocab_size: vocabulary size (existing vocabulary units in LM; should not include blank etc.)
            unk_score: score for unknown tokens
            final_eos_score: score for eos token after detected end of context phrase
            use_triton: allow using Triton implementation;
                None (default) means "auto" (used if available), True means forced mode
                (will crash if Triton is unavailable)
            uniform_weights: whether to use uniform weights for the context-biasing tree as in Icefall

        Returns:
            GPUBoostingTreeModel instance
        """
        logging.info(f"{cls.__name__}: reading boosting tree from {context_graph}")

        order2cnt, tbranches_list = cls._read_context_graph(context_graph=context_graph)

        # init suffix tree storage
        max_states = context_graph.num_nodes + 1  # + 1 for root state
        boosting_tree_np = BoostingTreeStorage(
            num_states_max=max_states,
            num_states=0,
            num_arcs=0,
            num_arcs_max=max_states * 2 + vocab_size * 2 + 1,
            unk_score=unk_score,
            final_eos_score=final_eos_score,
            vocab_size=vocab_size,
            max_order=max(order2cnt) + 1,
        )

        boosting_tree_np.uniform_weights = uniform_weights
        # convert context-biasing graph to np boosting tree
        tbranch_cur_order_i = 0
        cur_order = 1

        for tbranch in tbranches_list:

            if tbranch_cur_order_i == 0:
                boosting_tree_np._start_adding_tbranches_for_order(order=cur_order)
            tbranch_cur_order_i += 1

            # add tbranch
            boosting_tree_np._tbranches.append(tbranch)
            boosting_tree_np._tbranches_cnt += 1

            if tbranch_cur_order_i == order2cnt[cur_order]:
                boosting_tree_np._end_adding_tbranches_for_order(order=cur_order)
                logging.debug(f"Processed {order2cnt[cur_order]} n-grams of order {cur_order}")
                cur_order += 1
                tbranch_cur_order_i = 0

        assert tbranch_cur_order_i == 0
        boosting_tree_np.sanity_check()
        logging.debug(f"Loaded boosting model with {len(tbranches_list)} arcs")
        return GPUBoostingTreeModel.from_boosting_tree_np(boosting_tree_np=boosting_tree_np, use_triton=use_triton)

    @classmethod
    def from_boosting_tree_np(
        cls, boosting_tree_np: BoostingTreeStorage, use_triton: bool | None = None
    ) -> "GPUBoostingTreeModel":
        """
        Constructor from suffix tree storage.

        Args:
            suffix_tree_np: suffix tree
            use_triton: allow using Triton implementation;
                None (default) means "auto" (used if available), True means forced mode
                (will crash if Triton is unavailable)

        Returns:
            GPUBoostingTreeModel instance
        """
        model = GPUBoostingTreeModel(
            OmegaConf.structured(
                BoostingTreeConfig(
                    num_states=boosting_tree_np.num_states,
                    num_arcs=boosting_tree_np.num_arcs,
                    max_order=boosting_tree_np.max_order,
                    vocab_size=boosting_tree_np.vocab_size,
                    use_triton=use_triton,
                )
            )
        )
        model._init_from_suffix_tree_np(suffix_tree_np=boosting_tree_np)
        model._resolve_final()
        return model

    def advance(self, states: torch.Tensor, eos_id: Optional[int] = None) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Advance `states` [B]: return scores [B, V] and next states [B, V] for full vocab
        Args:
            states: batch of states
            eos_id: if not None, for eos symbol use final state weight

        Returns:
            tuple with next states and scores
        """
        if self.use_triton and states.device.type == "cuda":
            scores, next_states = self._advance_triton(states=states)
        else:
            scores, next_states = self._advance_pytorch(states=states)

        # replace eos_id score with maximum state weight to prevent from hallucinating in case of AED models (e.g. Canary)
        if eos_id is not None:
            # 1. replace eos score with maximum boosting value at each step
            scores[:, eos_id] = torch.clamp(torch.max(scores, dim=1).values, min=0.0)

            # 2. increase eos score after detected end of context phrase
            scores[:, eos_id] += self.get_final(states)

            next_states[:, eos_id] = states
        return scores, next_states

    def get_final(self, states: torch.Tensor) -> torch.Tensor:
        """
        Get final weights for states

        Args:
            states: batch of states

        Returns:
            tensor [B] with final weights for each state
        """

        return self.final_weights[states]

    @classmethod
    def dummy_boosting_tree(
        cls,
        vocab_size: int,
        use_triton: bool | None = None,
    ) -> "GPUBoostingTreeModel":
        """
        Constructs a trivial boosting tree with only one context phrase without scores.
        Useful for testing purposes (e.g., decoding).

        Returns:
            GPUBoostingTreeModel instance
        """

        context_graph_trivial = ContextGraph(context_score=0.0, depth_scaling=0.0)
        context_graph_trivial.build(token_ids=[[1]], phrases=["c"], scores=[0.0], uniform_weights=False)

        boosting_tree_trivial = GPUBoostingTreeModel.from_context_graph(
            context_graph=context_graph_trivial,
            vocab_size=vocab_size,
            unk_score=0.0,
            final_eos_score=0.0,
            use_triton=use_triton,
            uniform_weights=False,
        )
        return boosting_tree_trivial

    @classmethod
    def get_alternative_transcripts(
        cls, cfg: BoostingTreeModelConfig, tokenizer: TokenizerSpec, phrase: str
    ) -> list[list[int]]:
        """
        Get alternative transcriptions for a key phrase using BPE dropout
        """
        i = 1
        cur_step = 1
        transcripts_set = set()
        transcripts_list = [tokenizer.text_to_ids(phrase)]
        while i < cfg.num_of_transcriptions and cur_step < cfg.num_of_transcriptions * 5:
            cur_step += 1
            transcript = tokenizer.tokenizer.encode(phrase, enable_sampling=True, alpha=cfg.bpe_alpha, nbest_size=-1)
            transcript_text = tokenizer.ids_to_tokens(transcript)
            if transcript_text[0] == "▁":  # skip the case of empty first token
                continue
            transcript_text = " ".join(transcript_text)
            if transcript_text not in transcripts_set:
                transcripts_list.append(transcript)
                transcripts_set.add(transcript_text)
                i += 1
        return transcripts_list

    @classmethod
    def from_arpa(
        cls,
        lm_path: Path | str,
        vocab_size: int,
        normalize_unk: bool = True,
        use_triton: bool | None = None,
        token_offset: int = DEFAULT_TOKEN_OFFSET,
    ) -> "NGramGPULanguageModel":
        raise NeMoBaseException("Boosting tree cannot be loaded from ARPA file")

    @classmethod
    def from_config(cls, cfg: BoostingTreeModelConfig, tokenizer: TokenizerSpec) -> "GPUBoostingTreeModel":
        """
        Constructor boosting tree model from config file
        """
        # load boosting tree from already built model path
        if cfg.model_path is not None and os.path.exists(cfg.model_path):
            return cls.from_nemo(lm_path=cfg.model_path, vocab_size=tokenizer.vocab_size)

        # 1. read key phrases from file or list
        phrase_items_list: list[PhraseItem]
        if cfg.key_phrases_file is not None and bool(cfg.key_phrases_list or cfg.key_phrase_items_list):
            raise ValueError("Both file and phrases specified, use only one")
        elif cfg.key_phrases_file:
            with open(cfg.key_phrases_file, "r", encoding="utf-8") as f:
                phrase_items_list = [PhraseItem(line.strip(), cfg.source_lang) for line in f]
        elif cfg.key_phrases_list or cfg.key_phrase_items_list:
            phrase_items_list = []
            if cfg.key_phrases_list:
                phrase_items_list = [PhraseItem(phrase, cfg.source_lang) for phrase in cfg.key_phrases_list]
            if cfg.key_phrase_items_list:
                phrase_items_list += cfg.key_phrase_items_list
        else:
            raise ValueError("No key phrases file or list specified")

        # 2. tokenize key phrases
        phrases_dict = {}
        is_aggregate_tokenizer = isinstance(tokenizer, AggregateTokenizer)

        use_bpe_dropout = cfg.use_bpe_dropout
        if use_bpe_dropout:
            if is_aggregate_tokenizer:
                logging.warning(
                    "Aggregated tokenizer does not support BPE dropout, only one default transcription will be used..."
                )
                use_bpe_dropout = False
            spm.set_random_generator_seed(1234)  # fix random seed for reproducibility of BPE dropout

        for phrase_item in phrase_items_list:
            phrase = phrase_item.phrase
            if use_bpe_dropout:
                phrases_dict[phrase] = cls.get_alternative_transcripts(cfg, tokenizer, phrase)
            else:
                if is_aggregate_tokenizer:
                    phrases_dict[phrase] = tokenizer.text_to_ids(phrase, phrase_item.lang)
                else:
                    phrases_dict[phrase] = tokenizer.text_to_ids(phrase)

        # 3. build pythoncontext graph
        contexts, scores, phrases = [], [], []
        for phrase in phrases_dict:
            if use_bpe_dropout:
                for transcript in phrases_dict[phrase]:
                    contexts.append(transcript)
                    scores.append(round(cfg.score_per_phrase / len(phrase), 2))
                    phrases.append(phrase)
            else:
                contexts.append(phrases_dict[phrase])
                scores.append(round(cfg.score_per_phrase / len(phrase), 2))
                phrases.append(phrase)

        context_graph = ContextGraph(context_score=cfg.context_score, depth_scaling=cfg.depth_scaling)
        context_graph.build(token_ids=contexts, scores=scores, phrases=phrases, uniform_weights=cfg.uniform_weights)

        # 4. build GPU boosting tree model from python context graph
        boosting_tree_model = GPUBoostingTreeModel.from_context_graph(
            context_graph=context_graph,
            vocab_size=tokenizer.vocab_size,
            unk_score=cfg.unk_score,
            final_eos_score=cfg.final_eos_score,
            use_triton=cfg.use_triton,
            uniform_weights=cfg.uniform_weights,
        )

        # 5. save model
        if cfg.model_path is not None:
            boosting_tree_model.save_to(cfg.model_path)

        return boosting_tree_model



# Copyright (c) 2023, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Optional, Tuple, Union


@dataclass
class DiarizerComponentConfig:
    """Dataclass to imitate HydraConfig dict when accessing parameters."""

    def get(self, name: str, default: Optional[Any] = None):
        return getattr(self, name, default)

    def __iter__(self):
        for key in asdict(self):
            yield key

    def dict(self) -> Dict:
        return asdict(self)


@dataclass
class ASRDiarizerCTCDecoderParams:
    pretrained_language_model: Optional[str] = None  # KenLM model file: .arpa model file or .bin binary file.
    beam_width: int = 32
    alpha: float = 0.5
    beta: float = 2.5


@dataclass
class ASRRealigningLMParams:
    # Provide a KenLM language model in .arpa format.
    arpa_language_model: Optional[str] = None
    # Min number of words for the left context.
    min_number_of_words: int = 3
    # Max number of words for the right context.
    max_number_of_words: int = 10
    # The threshold for the difference between two log probability values from two hypotheses.
    logprob_diff_threshold: float = 1.2


@dataclass
class ASRDiarizerParams(DiarizerComponentConfig):
    # if True, speech segmentation for diarization is based on word-timestamps from ASR inference.
    asr_based_vad: bool = False
    # Threshold (in sec) that caps the gap between two words when generating VAD timestamps using ASR based VAD.
    asr_based_vad_threshold: float = 1.0
    # Batch size can be dependent on each ASR model. Default batch sizes are applied if set to null.
    asr_batch_size: Optional[int] = None
    # Native decoder delay. null is recommended to use the default values for each ASR model.
    decoder_delay_in_sec: Optional[float] = None
    # Offset to set a reference point from the start of the word. Recommended range of values is [-0.05  0.2].
    word_ts_anchor_offset: Optional[float] = None
    # Select which part of the word timestamp we want to use. The options are: 'start', 'end', 'mid'.
    word_ts_anchor_pos: str = "start"
    # Fix the word timestamp using VAD output. You must provide a VAD model to use this feature.
    fix_word_ts_with_VAD: bool = False
    # If True, use colored text to distinguish speakers in the output transcript.
    colored_text: bool = False
    # If True, the start and end time of each speaker turn is printed in the output transcript.
    print_time: bool = True
    # If True, the output transcript breaks the line to fix the line width (default is 90 chars)
    break_lines: bool = False


@dataclass
class ASRDiarizerConfig(DiarizerComponentConfig):
    model_path: Optional[str] = "stt_en_conformer_ctc_large"
    parameters: ASRDiarizerParams = field(default_factory=lambda: ASRDiarizerParams())
    ctc_decoder_parameters: ASRDiarizerCTCDecoderParams = field(default_factory=lambda: ASRDiarizerCTCDecoderParams())
    realigning_lm_parameters: ASRRealigningLMParams = field(default_factory=lambda: ASRRealigningLMParams())


@dataclass
class VADParams(DiarizerComponentConfig):
    window_length_in_sec: float = 0.15  # Window length in sec for VAD context input
    shift_length_in_sec: float = 0.01  # Shift length in sec for generate frame level VAD prediction
    smoothing: Union[str, bool] = "median"  # False or type of smoothing method (eg: median)
    overlap: float = 0.5  # Overlap ratio for overlapped mean/median smoothing filter
    onset: float = 0.1  # Onset threshold for detecting the beginning and end of a speech
    offset: float = 0.1  # Offset threshold for detecting the end of a speech
    pad_onset: float = 0.1  # Adding durations before each speech segment
    pad_offset: float = 0  # Adding durations after each speech segment
    min_duration_on: float = 0  # Threshold for short speech segment deletion
    min_duration_off: float = 0.2  # Threshold for small non_speech deletion
    filter_speech_first: bool = True


@dataclass
class VADConfig(DiarizerComponentConfig):
    model_path: str = "vad_multilingual_marblenet"  # .nemo local model path or pretrained VAD model name
    external_vad_manifest: Optional[str] = None
    parameters: VADParams = field(default_factory=lambda: VADParams())


@dataclass
class SpeakerEmbeddingsParams(DiarizerComponentConfig):
    # Window length(s) in sec (floating-point number). either a number or a list. ex) 1.5 or [1.5,1.0,0.5]
    window_length_in_sec: Tuple[float] = (1.5, 1.25, 1.0, 0.75, 0.5)
    # Shift length(s) in sec (floating-point number). either a number or a list. ex) 0.75 or [0.75,0.5,0.25]
    shift_length_in_sec: Tuple[float] = (0.75, 0.625, 0.5, 0.375, 0.25)
    # Weight for each scale. None (for single scale) or list with window/shift scale count. ex) [0.33,0.33,0.33]
    multiscale_weights: Tuple[float] = (1, 1, 1, 1, 1)
    # save speaker embeddings in pickle format. True if clustering result is used for other models, such as MSDD.
    save_embeddings: bool = True


@dataclass
class SpeakerEmbeddingsConfig(DiarizerComponentConfig):
    # .nemo local model path or pretrained model name (titanet_large, ecapa_tdnn or speakerverification_speakernet)
    model_path: Optional[str] = None
    parameters: SpeakerEmbeddingsParams = field(default_factory=lambda: SpeakerEmbeddingsParams())


@dataclass
class ClusteringParams(DiarizerComponentConfig):
    # If True, use num of speakers value provided in manifest file.
    oracle_num_speakers: bool = False
    # Max number of speakers for each recording. If an oracle number of speakers is passed, this value is ignored.
    max_num_speakers: int = 8
    # If the number of segments is lower than this number, enhanced speaker counting is activated.
    enhanced_count_thres: int = 80
    # Determines the range of p-value search: 0 < p <= max_rp_threshold.
    max_rp_threshold: float = 0.25
    # The higher the number, the more values will be examined with more time.
    sparse_search_volume: int = 30
    # If True, take a majority vote on multiple p-values to estimate the number of speakers.
    maj_vote_spk_count: bool = False


@dataclass
class ClusteringConfig(DiarizerComponentConfig):
    parameters: ClusteringParams = field(default_factory=lambda: ClusteringParams())


@dataclass
class MSDDParams(DiarizerComponentConfig):
    # If True, use speaker embedding model in checkpoint, else provided speaker embedding model in config will be used.
    use_speaker_model_from_ckpt: bool = True
    # Batch size for MSDD inference.
    infer_batch_size: int = 25
    # Sigmoid threshold for generating binarized speaker labels. The smaller the more generous on detecting overlaps.
    sigmoid_threshold: Tuple[float] = (0.7,)
    # If True, use oracle number of speaker and evaluate F1 score for the given speaker sequences. Default is False.
    seq_eval_mode: bool = False
    # If True, break the input audio clip to short sequences and calculate cluster average embeddings for inference.
    split_infer: bool = True
    # The length of split short sequence when split_infer is True.
    diar_window_length: int = 50
    # If the estimated number of speakers are larger than this number, overlap speech is not estimated.
    overlap_infer_spk_limit: int = 5


@dataclass
class MSDDConfig(DiarizerComponentConfig):
    model_path: Optional[str] = "diar_msdd_telephonic"
    parameters: MSDDParams = field(default_factory=lambda: MSDDParams())


@dataclass
class DiarizerConfig(DiarizerComponentConfig):
    manifest_filepath: Optional[str] = None
    out_dir: Optional[str] = None
    oracle_vad: bool = False  # If True, uses RTTM files provided in the manifest file to get VAD timestamps
    collar: float = 0.25  # Collar value for scoring
    ignore_overlap: bool = True  # Consider or ignore overlap segments while scoring
    vad: VADConfig = field(default_factory=lambda: VADConfig())
    speaker_embeddings: SpeakerEmbeddingsConfig = field(default_factory=lambda: SpeakerEmbeddingsConfig())
    clustering: ClusteringConfig = field(default_factory=lambda: ClusteringConfig())
    msdd_model: MSDDConfig = field(default_factory=lambda: MSDDConfig())
    asr: ASRDiarizerConfig = field(default_factory=lambda: ASRDiarizerConfig())


@dataclass
class NeuralDiarizerInferenceConfig(DiarizerComponentConfig):
    diarizer: DiarizerConfig = field(default_factory=lambda: DiarizerConfig())
    device: str = "cpu"
    verbose: bool = False
    batch_size: int = 64
    num_workers: int = 1
    sample_rate: int = 16000
    name: str = ""

    @classmethod
    def init_config(cls, diar_model_path: str, vad_model_path: str, map_location: str, verbose: bool):
        return NeuralDiarizerInferenceConfig(
            DiarizerConfig(
                vad=VADConfig(model_path=vad_model_path),
                msdd_model=MSDDConfig(model_path=diar_model_path),
            ),
            device=map_location,
            verbose=verbose,
        )
